@@ -1,8 +1,9 @@
 import importlib
 import sys
 
-from ptera import Probe
+from ptera import Probe, global_probe, probing
 
+from . import plugins
 from .analyze import explore
 from .debug import SnekTalkDb
 from .evaluator import Evaluator
