@@ -156,37 +156,92 @@ define(["vs/editor/editor.main", "scripts/repl"], (monaco, repl) => {
                 }
             )
 
+            function selectedWord(ed) {
+                // Send the selection, or the word under the cursor if
+                // there is no selection. I'm sure there's an actual
+                // function somewhere in monaco to do this, because this
+                // is how actions that need a selection seem to already
+                // work, but finding it seems more time consuming than
+                // implementing it.
+                const model = ed.getModel();
+                let sel = ed.getSelection();
+                if (sel.startLineNumber === sel.endLineNumber
+                    && sel.startColumn === sel.endColumn) {
+
+                    const pos = ed.getPosition();
+                    const word = model.getWordAtPosition(pos);
+                    sel = {
+                        startLineNumber: pos.lineNumber,
+                        startColumn: word.startColumn,
+                        endLineNumber: pos.lineNumber,
+                        endColumn: word.endColumn,
+                    }
+                }
+                return sel;
+            }
+
             this.editor.addAction({
-                id: 'add-probe',
+                id: 'sktk-probe',
                 label: 'Probe',
-                keybindings: [KM.CtrlCmd | KM.Alt | KC.KEY_P],
+                keybindings: [KM.chord(KM.CtrlCmd | KM.Alt | KC.KEY_P, KC.KEY_P)],
                 precondition: null,
                 keybindingContext: null,
                 contextMenuGroupId: '99_probe',
                 contextMenuOrder: 1,
                 run: ed => {
-                    // Send the selection, or the word under the cursor if
-                    // there is no selection. I'm sure there's an actual
-                    // function somewhere in monaco to do this, because this
-                    // is how actions that need a selection seem to already
-                    // work, but finding it seems more time consuming than
-                    // implementing it.
-                    const model = ed.getModel();
-                    let sel = ed.getSelection();
-                    if (sel.startLineNumber === sel.endLineNumber
-                        && sel.startColumn === sel.endColumn) {
+                    this.options.py.probe(selectedWord(ed));
+                }
+            });
 
-                        const pos = ed.getPosition();
-                        const word = model.getWordAtPosition(pos);
-                        sel = {
-                            startLineNumber: pos.lineNumber,
-                            startColumn: word.startColumn,
-                            endLineNumber: pos.lineNumber,
-                            endColumn: word.endColumn,
-                        }
-                    }
+            this.editor.addAction({
+                id: 'sktk-local-probe',
+                label: 'Local probe',
+                keybindings: [KM.chord(KM.CtrlCmd | KM.Alt | KC.KEY_P, KC.KEY_L)],
+                precondition: null,
+                keybindingContext: null,
+                contextMenuGroupId: '99_probe',
+                contextMenuOrder: 1,
+                run: ed => {
+                    this.options.py.local_probe(selectedWord(ed));
+                }
+            });
 
-                    this.options.py.probe(sel);
+            this.editor.addAction({
+                id: 'sktk-explore',
+                label: 'Explore',
+                keybindings: [KM.chord(KM.CtrlCmd | KM.Alt | KC.KEY_P, KC.KEY_E)],
+                precondition: null,
+                keybindingContext: null,
+                contextMenuGroupId: '99_probe',
+                contextMenuOrder: 1,
+                run: ed => {
+                    this.options.py.explore(selectedWord(ed));
+                }
+            });
+
+            this.editor.addAction({
+                id: 'sktk-local-explore',
+                label: 'Local explore',
+                keybindings: [KM.chord(KM.CtrlCmd | KM.Alt | KC.KEY_P, KC.KEY_X)],
+                precondition: null,
+                keybindingContext: null,
+                contextMenuGroupId: '99_probe',
+                contextMenuOrder: 1,
+                run: ed => {
+                    this.options.py.local_explore(selectedWord(ed));
+                }
+            });
+
+            this.editor.addAction({
+                id: 'sktk-accumulate',
+                label: 'Accumulate',
+                keybindings: [KM.chord(KM.CtrlCmd | KM.Alt | KC.KEY_P, KC.KEY_A)],
+                precondition: null,
+                keybindingContext: null,
+                contextMenuGroupId: '99_probe',
+                contextMenuOrder: 1,
+                run: ed => {
+                    this.options.py.accumulate(selectedWord(ed));
                 }
             });
         }
