@@ -40,7 +40,9 @@ function isElementInViewport(el) {
 }
 
 require.config({ paths: { 'vs': '/lib/vs' }});
-define(["vs/editor/editor.main"], (monaco) => {
+define(["vs/editor/editor.main", "lib/mousetrap.min"], (monaco, Mousetrap) => {
+
+Mousetrap.prototype.stopCallback = () => false;
 
 const KM = monaco.KeyMod;
 const KC = monaco.KeyCode;
@@ -49,7 +51,37 @@ let mainRepl = null;
 let evalIdGen = 0;
 
 
+class PriorityList {
+
+    constructor() {
+        this.handlers = [];
+    }
+
+    add(priority, data) {
+        this.handlers.push([priority, data]);
+        this.handlers.sort((a, b) => b[0] - a[0])
+    }
+
+    push(data) {
+        this.add(this.handlers.length + 1, data);
+    }
+
+    run(...args) {
+        for (const data of this) {
+            data(...args);
+        }
+    }
+
+    *[Symbol.iterator]() {
+        for (const [priority, data] of this.handlers) {
+            yield data;
+        }
+    }
+}
+
+
 class PopupNav {
+
     constructor(options) {
         this.name = options.name;
         this.default = options.default;
@@ -168,20 +200,136 @@ class PopupNav {
 }
 
 
+class PopupSet {
+
+    constructor(popups) {
+        this.active = null;
+        this.popups = Object.assign({}, popups || {});
+        this.lib = null;
+    }
+
+    setup(repl) {
+        this.repl = repl;
+        this.lib = this.repl.lib;
+    }
+
+    context() {
+        return this.repl.editor.getValue();
+    }
+
+    async toggle(name) {
+        this.active = null;
+        for (const popupName in this.popups) {
+            const pop = this.popups[popupName];
+            if (name == popupName && !pop.visible) {
+                const entries = await this.lib.populate_popup(
+                    name,
+                    this.context(),
+                );
+                if (entries !== null) {
+                    pop.setEntries(entries);
+                }
+                pop.show();
+                if (!pop.entries.length) {
+                    pop.setCursor(0);
+                }
+                this.active = pop;
+            }
+            else {
+                pop.hide();
+            }
+        }
+        this.repl.popupActive.set(this.active !== null);
+    }
+
+}
+
+
+class PinPane {
+
+    constructor(element) {
+        this.repl = null;
+        this.element = element;
+    }
+
+    setup(repl) {
+        this.repl = repl;
+        this.repl.events.click.add(
+            200,
+            {
+                modifiers: {"cmdCtrl": true, "altKey": true, "shiftKey": false},
+                attr: "pinnable",
+                handler: this.clickHandler.bind(this),
+            }
+        )
+        this.repl.events.printbox.push(prbox => {
+            prbox.setAttribute("pinnable", "");
+        })
+    }
+
+    clickHandler(evt, target) {
+        this.pin(target);
+        return true;
+    }
+
+    pin(elem) {
+        if (!elem.$referenceParent) {
+            elem.$referenceParent = elem.parentElement;
+        }
+        let parent = elem.$referenceParent;
+
+        let make_button = () => {
+            let pinbutton = document.createElement("span");
+            pinbutton.className = "snek-pinbutton";
+            pinbutton.innerHTML = `<svg width="24" height="24" fill="#aaa" transform="scale(0.75, 0.75)" viewBox="0 0 1000 1000" xmlns="http://www.w3.org/2000/svg"><path d=" M 725 88C 746 88 762 104 763 125C 763 125 763 188 763 188C 763 242 725 287 675 287C 675 287 641 287 641 287C 641 287 659 466 659 466C 690 473 717 489 734 513C 756 544 763 583 763 625C 762 646 746 662 725 663C 725 663 600 663 600 663C 600 663 563 663 563 663C 563 663 437 663 437 663C 437 663 400 663 400 663C 400 663 275 663 275 663C 254 662 238 646 238 625C 238 583 244 544 266 513C 283 489 310 473 341 466C 341 466 359 287 359 287C 359 287 325 287 325 287C 300 287 279 276 264 261C 249 246 238 225 238 200C 238 158 238 167 238 125C 238 104 254 88 275 88C 275 88 725 88 725 88M 563 710C 563 710 563 850 563 850C 563 856 561 862 559 867C 559 867 534 917 534 917C 527 929 514 937 500 937C 486 937 473 929 466 917C 466 917 441 867 441 867C 439 862 438 856 437 850C 437 850 437 710 437 710C 437 710 563 710 563 710"/></svg>`
+            pinbutton.onclick = event => {
+                event.preventDefault();
+                event.stopPropagation();
+                this.pin(elem);
+            }
+            return pinbutton;
+        }
+
+        if (!elem.$pinned) {
+            let wrapper = document.createElement("div");
+            if (!elem.$onclick) {
+                elem.$onclick = elem.onclick;
+            }
+            let pinbutton = make_button();
+            parent.replaceChild(pinbutton, elem);
+            this.element.appendChild(wrapper);
+            wrapper.appendChild(elem);
+            elem.$wrapper = wrapper;
+            elem.$pinned = pinbutton;
+        }
+        else {
+            this.element.removeChild(elem.$wrapper);
+            parent.replaceChild(elem, elem.$pinned);
+            elem.$pinned = false;
+        }
+    }
+}
+
+
 class Repl {
 
     constructor(target) {
+        this.events = {
+            click: new PriorityList(),
+            printbox: new PriorityList(),
+        };
         this.container = target;
         this.options = {};
         this.lib = {};
         this.pane = target.querySelector(".snek-pane");
         this.outerPane = target.querySelector(".snek-outer-pane");
-        this.pinpane = target.querySelector(".snek-pin-pane");
+        this.pinpane = new PinPane(target.querySelector(".snek-pin-pane"));
+        this.pinpane.setup(this);
         this.inputOuter = target.querySelector(".snek-input-box");
         this.inputBox = target.querySelector(".snek-input");
         this.inputMode = target.querySelector(".snek-input-mode");
         this.inputMode.onclick = () => setTimeout(() => {
-            this.togglePopup("interactors");
+            this.popups.toggle("interactors");
             this.editor.focus();
         }, 0);
         this.statusBar = target.querySelector(".snek-status-bar");
@@ -189,9 +337,8 @@ class Repl {
         this.nav = target.querySelector(".snek-nav");
         this.statusHistory = document.createElement("div");
         this.statusHistory.className = "snek-status-history";
-        this._setupEditor(this.inputBox);
 
-        this.popups = {
+        this.popups = new PopupSet({
             history: new PopupNav({
                 name: "history",
                 anchor: this.inputOuter,
@@ -199,7 +346,7 @@ class Repl {
                     this.editor.setValue(entries.map(x => x.text).join("\n"));
                     const nlines = this.editor.getModel().getLineCount();
                     this.editor.setPosition({lineNumber: nlines, column: 1000000});
-                    this.togglePopup(null);
+                    this.popups.toggle(null);
                     this.editor.focus();
                 },
                 class: "snek-popup-nav-history",
@@ -213,55 +360,82 @@ class Repl {
                         command: "submit",
                         expr: `/attach ${entries[0].text}`,
                     });
-                    this.togglePopup(null);
+                    this.popups.toggle(null);
                     this.editor.focus();
                 },
                 class: "snek-popup-nav-interactors",
                 default: "No interactors",
             })
-        };
-        this.togglePopup(null);
+        });
+        this.$$setupEditor(this.inputBox);
+
+        this.popups.setup(this);
+        this.popups.toggle(null);
 
-        this.historySelection = 0;
-        this.filter = null;
-        this.filteredHistory = [""];
-        this.history = [""];
         this.expectedContent = null;
-        this.historyPopup = null;
 
-        target.onclick = this.$globalClickEvent.bind(this);
-        window.onkeydown = this.$globalKDEvent.bind(this);
+        target.onclick = this.$$globalClickEvent.bind(this);
+        this.$$addKeyboardHandlers();
         window.$$SKTK = this.sktk.bind(this);
         window.snektalk = this;
         this.$currid = 0;
         this.$responseMap = {};
+
+        this.$$addClickHandlers();
+
         exports.mainRepl = this;
     }
 
-    $globalKDEvent(evt) {
-        let cmdctrl = isMac ? (evt.metaKey && !evt.ctrlKey) : evt.ctrlKey;
+    $$addClickHandlers() {
+        // Default behavior
+        this.events.click.add(
+            0,
+            {
+                modifiers: {},
+                attr: null,
+                handler(evt, target) {
+                    if (target.onclick !== null
+                        || target.onmousedown !== null
+                        || target.ondblclick !== null
+                        || target.tagName === "A") {
+                        // Stop if non-default behavior or link
 
-        // Cmd+P => focus repl
-        if (cmdctrl && !evt.altKey && evt.key === "p") {
+                        return true
+                    }
+                    else {
+                        return false
+                    }
+                }
+            }
+        );
 
-            evt.preventDefault();
-            evt.stopPropagation();
+        // Click on object
+        this.events.click.add(
+            100,
+            {
+                modifiers: {"cmdCtrl": true, "altKey": false, "shiftKey": false},
+                attr: "objid",
+                handler(evt, target) {
+                    this.sktk(parseInt(target.getAttribute("objid")));
+                    return true;
+                }
+            }
+        );
+    }
 
-            this.editor.focus();
-        }
-        // Cmd+B => navigate to next visible editor
-        else if (cmdctrl && !evt.altKey && evt.key === "b") {
+    //////////////////
+    // Initializers //
+    //////////////////
 
-            evt.preventDefault();
-            evt.stopPropagation();
-    
+    $$addKeyboardHandlers() {
+        const cycleEditor = direction => {
             // Find alive and visible editors
             let editors = Array
             .from(this.container.querySelectorAll(".snek-editor-cyclable"))
             .filter(container => isElementInViewport(container))
             .map(container => container.$editor);
 
-            if (!evt.shiftKey) {
+            if (direction < 0) {
                 editors.reverse();
             }
 
@@ -279,31 +453,48 @@ class Repl {
                 editors[0].focus();
             }
         }
+
+        this.bindKey("mod+p", () => {
+            this.editor.focus();
+            return false;
+        });
+
+        this.bindKey("mod+b", () => {
+            cycleEditor(-1);
+            return false;
+        });
+
+        this.bindKey("mod+shift+b", () => {
+            cycleEditor(1);
+            return false;
+        });
     }
 
-    $globalClickEvent(evt) {
-        let cmdctrl = isMac ? (evt.metaKey && !evt.ctrlKey) : evt.ctrlKey;
+    $$globalClickEvent(evt) {
+        evt.cmdCtrl = isMac ? (evt.metaKey && !evt.ctrlKey) : evt.ctrlKey;
 
         if (evt.detail === 2) {
             // Double-click
             return;
         }
         let target = evt.target;
+
+        outer_loop:
         while (target) {
-            if (target.onclick !== null
-                || target.onmousedown !== null
-                || target.ondblclick !== null
-                || target.tagName === "A") {
-                // Stop if non-default behavior or link
-                break;
-            }
-            else if (cmdctrl && evt.altKey && !evt.shiftKey && target.getAttribute("pinnable") !== null) {
-                this.pin(target);
-                break;
-            }
-            else if (cmdctrl && !evt.altKey && !evt.shiftKey && target.getAttribute("objid") !== null) {
-                this.sktk(parseInt(target.getAttribute("objid")));
-                break;
+
+            handler_loop:
+            for (let {modifiers, attr, handler} of this.events.click) {
+                for (let key in modifiers) {
+                    if (evt[key] != modifiers[key]) {
+                        continue handler_loop;
+                    }
+                }
+                if (attr && target.getAttribute(attr) === null) {
+                    continue handler_loop;
+                }
+                if (handler.call(this, evt, target)) {
+                    break outer_loop;
+                }
             }
             target = target.parentElement;
         }
@@ -314,7 +505,200 @@ class Repl {
             }
             target = target.parentElement;
         }
-        this.togglePopup(null);
+        this.popups.toggle(null);
+    }
+
+    $$event_updateHeight() {
+        if (!this.ignoreEvent) {
+            const contentHeight = Math.min(
+                this.options.max_height || 500,
+                this.editor.getContentHeight()
+            );
+            this.inputBox.style.height = `${contentHeight}px`;
+            try {
+                this.ignoreEvent = true;
+                this.editor.layout({ width: this.inputBox.offsetWidth, height: contentHeight });
+            } finally {
+                this.ignoreEvent = false;
+            }
+        }
+    }
+
+    $$setupEditor(target) {
+        target.style.height = "19px";
+        let editor = monaco.editor.create(target, {
+            value: "",
+            language: 'python',
+            lineNumbers: false,
+            minimap: {enabled: false},
+            scrollBeyondLastLine: false,
+            overviewRulerLanes: 0,
+            folding: false,
+            renderLineHighlight: "none",
+            lineDecorationsWidth: 0,
+            suggestOnTriggerCharacters: false,
+            wordBasedSuggestions: false,
+        });
+        this.editor = editor;
+        target.classList.add("snek-editor-cyclable");
+        target.$editor = this.editor;
+        this.editor.onDidContentSizeChange(this.$$event_updateHeight.bind(this));
+
+        this.popupActive = this.editor.createContextKey("popupActive", false);
+        this.atBeginning = this.editor.createContextKey("atBeginning", true);
+        this.atEnd = this.editor.createContextKey("atEnd", true);
+        this.multiline = this.editor.createContextKey("multiline", false);
+
+        const setMultiline = value => {
+            this.multiline.set(value);
+            if (value) {
+                this.inputBox.classList.add("multiline");
+            }
+            else {
+                this.inputBox.classList.remove("multiline");
+            }
+        }
+
+        this.editor.onDidChangeCursorPosition(() => {
+            let position = editor.getPosition();
+            let lineno = position.lineNumber;
+            let total = editor.getModel().getLineCount();
+            this.atBeginning.set(lineno == 1);
+            this.atEnd.set(lineno == total);
+        });
+
+        this.editor.getModel().onDidChangeContent(async () => {
+            let total = editor.getModel().getLineCount();
+            setMultiline(
+                total > 1
+                || editor.getModel().getLineContent(1).match(/[\(\[\{:]$|^@/)
+            )
+            let pop = this.popups.active;
+            if (pop !== null) {
+                pop.setEntries(
+                    await this.lib.populate_popup(pop.name, this.editor.getValue())
+                );
+            }
+        });
+
+        let _submit = () => {
+            let val = editor.getValue();
+            editor.setValue("");
+            this.send({
+                command: "submit",
+                expr: val,
+            });
+            this.popups.toggle(null);
+            // The flex-direction on the outer pane is reversed, so
+            // 0 scrolls it to the bottom. Handy.
+            this.outerPane.scrollTop = 0;
+        }
+
+        editor.addCommand(
+            KC.Enter,
+            _submit,
+            "!multiline && !popupActive"
+        );
+
+        editor.addCommand(
+            KM.WinCtrl | KC.Enter,
+            _submit
+        );
+
+        editor.addCommand(
+            KM.WinCtrl | KC.KEY_L,
+            () => { this.pane.innerHTML = ""; }
+        );
+
+        editor.addCommand(
+            KM.WinCtrl | KC.KEY_R,
+            () => this.popups.toggle("history")
+        );
+
+        editor.addCommand(
+            KC.UpArrow,
+            () => { this.navigateHistory(-1); },
+            "atBeginning && !popupActive"
+        );
+
+        editor.addCommand(
+            KC.DownArrow,
+            () => { this.navigateHistory(1); },
+            "atEnd && !popupActive"
+        );
+
+        // Bindings when popup is active
+
+        editor.addCommand(
+            KC.Enter,
+            () => this.popups.active.submit(),
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KC.Escape,
+            () => this.popups.toggle(null),
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KC.UpArrow,
+            () => { this.popups.active.goUp(); },
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KC.DownArrow,
+            () => { this.popups.active.goDown(); },
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KM.Shift | KC.UpArrow,
+            () => { this.popups.active.goUp(true); },
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KM.Shift | KC.DownArrow,
+            () => { this.popups.active.goDown(true); },
+            "popupActive"
+        );
+
+        // Misc
+
+        editor.addCommand(
+            KM.Alt | KC.US_SEMICOLON,
+            () => { this.editor.trigger(null, "editor.action.commentLine"); },
+        );
+
+        editor.addCommand(
+            KM.chord(KM.CtrlCmd | KC.KEY_K, KM.CtrlCmd | KC.KEY_K),
+            () => {
+                this.editor.setValue("");
+                this.lib.stop();
+            }
+        );
+
+        if (isMac) {
+            editor.addCommand(
+                KM.WinCtrl | KC.KEY_C,
+                () => {
+                    this.editor.setValue("");
+                    this.lib.stop();
+                }
+            );
+        }
+
+        this.editor = editor;
+    }
+
+    /////////////
+    // Methods //
+    /////////////
+
+    bindKey(key, fn) {
+        return Mousetrap.bind(key, fn);
     }
 
     get_external(id) {
@@ -377,258 +761,6 @@ class Repl {
         }
     }
 
-    pin(elem) {
-        if (!elem.$referenceParent) {
-            elem.$referenceParent = elem.parentElement;
-        }
-        let parent = elem.$referenceParent;
-
-        let make_button = () => {
-            let pinbutton = document.createElement("span");
-            pinbutton.className = "snek-pinbutton";
-            pinbutton.innerHTML = `<svg width="24" height="24" fill="#aaa" transform="scale(0.75, 0.75)" viewBox="0 0 1000 1000" xmlns="http://www.w3.org/2000/svg"><path d=" M 725 88C 746 88 762 104 763 125C 763 125 763 188 763 188C 763 242 725 287 675 287C 675 287 641 287 641 287C 641 287 659 466 659 466C 690 473 717 489 734 513C 756 544 763 583 763 625C 762 646 746 662 725 663C 725 663 600 663 600 663C 600 663 563 663 563 663C 563 663 437 663 437 663C 437 663 400 663 400 663C 400 663 275 663 275 663C 254 662 238 646 238 625C 238 583 244 544 266 513C 283 489 310 473 341 466C 341 466 359 287 359 287C 359 287 325 287 325 287C 300 287 279 276 264 261C 249 246 238 225 238 200C 238 158 238 167 238 125C 238 104 254 88 275 88C 275 88 725 88 725 88M 563 710C 563 710 563 850 563 850C 563 856 561 862 559 867C 559 867 534 917 534 917C 527 929 514 937 500 937C 486 937 473 929 466 917C 466 917 441 867 441 867C 439 862 438 856 437 850C 437 850 437 710 437 710C 437 710 563 710 563 710"/></svg>`
-            pinbutton.onclick = event => {
-                event.preventDefault();
-                event.stopPropagation();
-                this.pin(elem);
-            }
-            return pinbutton;    
-        }
-
-        if (!elem.$pinned) {
-            let wrapper = document.createElement("div");
-            if (!elem.$onclick) {
-                elem.$onclick = elem.onclick;
-            }
-            let pinbutton = make_button();
-            parent.replaceChild(pinbutton, elem);
-            this.pinpane.appendChild(wrapper);
-            wrapper.appendChild(elem);
-            elem.$wrapper = wrapper;
-            elem.$pinned = pinbutton;
-        }
-        else {
-            this.pinpane.removeChild(elem.$wrapper);
-            parent.replaceChild(elem, elem.$pinned);
-            elem.$pinned = false;
-        }
-    }
-
-    event_updateHeight() {
-        if (!this.ignoreEvent) {
-            const contentHeight = Math.min(
-                this.options.max_height || 500,
-                this.editor.getContentHeight()
-            );
-            this.inputBox.style.height = `${contentHeight}px`;
-            try {
-                this.ignoreEvent = true;
-                this.editor.layout({ width: this.inputBox.offsetWidth, height: contentHeight });
-            } finally {
-                this.ignoreEvent = false;
-            }
-        }
-    }
-
-    _setupEditor(target) {
-        target.style.height = "19px";
-        let editor = monaco.editor.create(target, {
-            value: "",
-            language: 'python',
-            lineNumbers: false,
-            minimap: {enabled: false},
-            scrollBeyondLastLine: false,
-            overviewRulerLanes: 0,
-            folding: false,
-            renderLineHighlight: "none",
-            lineDecorationsWidth: 0,
-            suggestOnTriggerCharacters: false,
-            wordBasedSuggestions: false,
-        });
-        this.editor = editor;
-        target.classList.add("snek-editor-cyclable");
-        target.$editor = this.editor;
-        this.editor.onDidContentSizeChange(this.event_updateHeight.bind(this));
-
-        this.popupActive = this.editor.createContextKey("popupActive", false);
-        this.atBeginning = this.editor.createContextKey("atBeginning", true);
-        this.atEnd = this.editor.createContextKey("atEnd", true);
-        this.multiline = this.editor.createContextKey("multiline", false);
-
-        const setMultiline = value => {
-            this.multiline.set(value);
-            if (value) {
-                this.inputBox.classList.add("multiline");
-            }
-            else {
-                this.inputBox.classList.remove("multiline");
-            }
-        }
-
-        this.editor.onDidChangeCursorPosition(() => {
-            let position = editor.getPosition();
-            let lineno = position.lineNumber;
-            let total = editor.getModel().getLineCount();
-            this.atBeginning.set(lineno == 1);
-            this.atEnd.set(lineno == total);
-        });
-
-        this.editor.getModel().onDidChangeContent(async () => {
-            let total = editor.getModel().getLineCount();
-            setMultiline(
-                total > 1
-                || editor.getModel().getLineContent(1).match(/[\(\[\{:]$|^@/)
-            )
-            let pop = this.activePopup;
-            if (pop !== null) {
-                pop.setEntries(
-                    await this.lib.populate_popup(pop.name, this.editor.getValue())
-                );
-            }
-        });
-
-        let _submit = () => {
-            let val = editor.getValue();
-            editor.setValue("");
-            this.send({
-                command: "submit",
-                expr: val,
-            });
-            if (this.history[1] !== val && val !== "") {
-                this.history[0] = val;
-                this.history.unshift("");
-            }
-            else {
-                this.history[0] = "";
-            }
-            this.filter = null;
-            this.togglePopup(null);
-            // The flex-direction on the outer pane is reversed, so
-            // 0 scrolls it to the bottom. Handy.
-            this.outerPane.scrollTop = 0;
-        }
-
-        editor.addCommand(
-            KC.Enter,
-            _submit,
-            "!multiline && !popupActive"
-        );
-
-        editor.addCommand(
-            KM.WinCtrl | KC.Enter,
-            _submit
-        );
-
-        editor.addCommand(
-            KM.WinCtrl | KC.KEY_L,
-            () => { this.pane.innerHTML = ""; }
-        );
-
-        editor.addCommand(
-            KM.WinCtrl | KC.KEY_R,
-            () => this.togglePopup("history")
-        );
-
-        editor.addCommand(
-            KC.UpArrow,
-            () => { this.navigateHistory(-1); },
-            "atBeginning && !popupActive"
-        );
-
-        editor.addCommand(
-            KC.DownArrow,
-            () => { this.navigateHistory(1); },
-            "atEnd && !popupActive"
-        );
-
-        // Bindings when popup is active
-
-        editor.addCommand(
-            KC.Enter,
-            () => this.activePopup.submit(),
-            "popupActive"
-        );
-
-        editor.addCommand(
-            KC.Escape,
-            () => this.togglePopup(null),
-            "popupActive"
-        );
-
-        editor.addCommand(
-            KC.UpArrow,
-            () => { this.activePopup.goUp(); },
-            "popupActive"
-        );
-
-        editor.addCommand(
-            KC.DownArrow,
-            () => { this.activePopup.goDown(); },
-            "popupActive"
-        );
-
-        editor.addCommand(
-            KM.Shift | KC.UpArrow,
-            () => { this.activePopup.goUp(true); },
-            "popupActive"
-        );
-
-        editor.addCommand(
-            KM.Shift | KC.DownArrow,
-            () => { this.activePopup.goDown(true); },
-            "popupActive"
-        );
-
-        // Misc
-
-        editor.addCommand(
-            KM.Alt | KC.US_SEMICOLON,
-            () => { this.editor.trigger(null, "editor.action.commentLine"); },
-        );
-
-        editor.addCommand(
-            KM.chord(KM.CtrlCmd | KC.KEY_K, KM.CtrlCmd | KC.KEY_K),
-            () => {
-                this.editor.setValue("");
-                this.lib.stop();
-            }
-        );
-
-        if (isMac) {
-            editor.addCommand(
-                KM.WinCtrl | KC.KEY_C,
-                () => {
-                    this.editor.setValue("");
-                    this.lib.stop();
-                }
-            );
-        }
-
-        this.editor = editor;
-    }
-
-    async togglePopup(name) {
-        this.activePopup = null;
-        for (const popupName in this.popups) {
-            const pop = this.popups[popupName];
-            if (name == popupName && !pop.visible) {
-                const entries = await this.lib.populate_popup(name, this.editor.getValue());
-                if (entries !== null) {
-                    pop.setEntries(entries);
-                }
-                pop.show();
-                if (!pop.entries.length) {
-                    pop.setCursor(0);
-                }
-                this.activePopup = pop;
-            }
-            else {
-                pop.hide();
-            }
-        }
-        this.popupActive.set(this.activePopup !== null);
-    }
-
     async navigateHistory(delta) {
         const text = (await this.lib.history_navigate(delta, this.editor.getValue()));
         if (text !== null) {
@@ -652,7 +784,7 @@ class Repl {
         return rval;
     }
 
-    read_only_editor(text, language) {
+    readOnlyEditor(text, language) {
         let elem = document.createElement("div");
         elem.style.width = this.pane.offsetWidth - 100;
 
@@ -663,44 +795,6 @@ class Repl {
         return elem;
     }
 
-    find_method(prefix, name, dflt) {
-        let method_name = `${prefix}_${name}`;
-        let method = this[method_name] || dflt;
-        return method && method.bind(this);
-    }
-
-    send(data) {
-        this.find_method("send", data.command, this.send_default)(data);
-    }
-
-    send_submit(data) {
-        let cmd = /^\/([^ ]+)( .*)?/.exec(data.expr);
-        let method = cmd && this.find_method("cmd", cmd[1], null);
-        if (method !== null) {
-            method(cmd[0], cmd[2]);
-        }
-        else {
-            this.send_default(data);
-        }
-    }
-
-    send_default(data) {
-        if (this.closed) {
-            this.setStatus({
-                type: "error",
-                value: "operation failed because the connection is closed",
-            });
-        }
-        else {
-            this.socket.send(JSON.stringify(data));
-        }
-    }
-
-    cmd_status(fullexpr, arg) {
-        this.recv_echo({value: fullexpr});
-        this.append(this.statusHistory, "plain");
-    }
-
     append(elem, type, target) {
         let wrapper = document.createElement("div");
         wrapper.className = "snek-line snek-t-" + type;
@@ -737,6 +831,84 @@ class Repl {
         );
     }
 
+    find_method(prefix, name, dflt) {
+        let method_name = `${prefix}_${name}`;
+        let method = this[method_name] || dflt;
+        return method && method.bind(this);
+    }
+
+    send(data) {
+        this.find_method("send", data.command, this.send_default)(data);
+    }
+
+    connect() {
+        let port = window.location.port;
+        let socket = new WebSocket(`ws://localhost:${port}/sktk?session=main`);
+
+        socket.addEventListener('message', event => {
+            let data = JSON.parse(event.data);
+            this.find_method("recv", data.command, this.recv_bad)(data)
+        });
+
+        socket.addEventListener('error', event => {
+            this.closed = true;
+            this.setStatus({
+                type: "error",
+                value: "a connection error occurred",
+            });
+        });
+
+        socket.addEventListener('close', event => {
+            this.closed = true;
+            this.setStatus({
+                type: "normal",
+                value: "the connection was closed",
+            });
+        });
+
+        this.socket = socket;
+    }
+
+    //////////////
+    // Commands //
+    //////////////
+
+    cmd_status(fullexpr, arg) {
+        this.recv_echo({value: fullexpr});
+        this.append(this.statusHistory, "plain");
+    }
+
+    /////////////
+    // Senders //
+    /////////////
+
+    send_submit(data) {
+        let cmd = /^\/([^ ]+)( .*)?/.exec(data.expr);
+        let method = cmd && this.find_method("cmd", cmd[1], null);
+        if (method !== null) {
+            method(cmd[0], cmd[2]);
+        }
+        else {
+            this.send_default(data);
+        }
+    }
+
+    send_default(data) {
+        if (this.closed) {
+            this.setStatus({
+                type: "error",
+                value: "operation failed because the connection is closed",
+            });
+        }
+        else {
+            this.socket.send(JSON.stringify(data));
+        }
+    }
+
+    ///////////////////
+    // Message types //
+    ///////////////////
+
     recv_resource(data) {
         let elem = this.reify(data.value);
         document.head.appendChild(elem);
@@ -749,7 +921,7 @@ class Repl {
         if (prbox === null) {
             prbox = document.createElement("div");
             prbox.id = "pr-eval-" + evalid;
-            prbox.setAttribute("pinnable", "");
+            this.events.printbox.run(prbox);
             this.pane.appendChild(prbox);
         }
         if (data.type === "statement") { }
@@ -786,7 +958,7 @@ class Repl {
     }
 
     recv_echo(data) {
-        let ed = this.read_only_editor(data.value.trimEnd(), data.language || "python");
+        let ed = this.readOnlyEditor(data.value.trimEnd(), data.language || "python");
         let elem = document.createElement("div");
         elem.appendChild(ed);
         this.append(elem, "echo");
@@ -831,10 +1003,6 @@ class Repl {
         this.inputMode.innerHTML = data.html;
     }
 
-    recv_add_history(data) {
-        this.history.push.apply(this.history, data["history"].reverse());
-    }
-
     recv_set_lib(data) {
         for (let key in data.lib) {
             this.lib[key] = this.get_external(data.lib[key]);
@@ -856,42 +1024,8 @@ class Repl {
     recv_bad(data) {
         console.error("Received an unknown command:", data.command);
     }
-
-    connect() {
-        let port = window.location.port;
-        let socket = new WebSocket(`ws://localhost:${port}/sktk?session=main`);
-
-        socket.addEventListener('message', event => {
-            let data = JSON.parse(event.data);
-            this.find_method("recv", data.command, this.recv_bad)(data)
-        });
-
-        socket.addEventListener('error', event => {
-            this.closed = true;
-            this.setStatus({
-                type: "error",
-                value: "a connection error occurred",
-            });
-        });
-
-        socket.addEventListener('close', event => {
-            this.closed = true;
-            this.setStatus({
-                type: "normal",
-                value: "the connection was closed",
-            });
-        });
-
-        this.socket = socket;
-    }
-
-    getEditor() {
-        return editor;
-    }
-
 }
 
-
 let exports = {
     mainRepl,
     Repl,
