@@ -102,7 +102,11 @@ def cli():
         **server_args,
     )
 
-    opts = SimpleNamespace(module=module, script=script, rest=argv,)
+    opts = SimpleNamespace(
+        module=module,
+        script=script,
+        rest=argv,
+    )
 
     mod, run = find_runner(opts, pattern)
     return mod, run, sess, thread
