@@ -30,7 +30,9 @@ def status_logger(sess):
             return
         else:
             sess.queue(
-                command="status", type="normal", value=str(event),
+                command="status",
+                type="normal",
+                value=str(event),
             )
 
     return log
