@@ -97,7 +97,8 @@ class NamedThreads:
             with session.set_context():
                 with new_evalid():
                     session.queue_result(
-                        H.div("Starting thread ", H.strong(word)), type="info",
+                        H.div("Starting thread ", H.strong(word)),
+                        type="info",
                     )
                     try:
                         fn()
